/* com/syncbuilder/app/BasicAppSkeleton.java:
*
* Copyright (C) 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: BasicAppSkeleton.java,v 1.7 1999/03/21 18:28:55 christ Exp $ */

package com.syncbuilder.app;


import com.syncbuilder.util.getopt.*;
import com.syncbuilder.util.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.IOException;
import java.io.PrintStream;


/**
 * This class is a skeleton for any command-line application. It uses an extremely powerful
 * command-line parser (derived from GNU getopt), and it can control the flow of the entire
 * initialization sequence of the app.<p>
 * It can also turn almost any command-line app into a simple GUI app. Every application which
 * you create with this skeleton will contain two standard options:<br>
 * <code>-gui</code> will turn on the GUI interface.<br>
 * <code>-help</code> will display usage information.
 *
 * @author Tilo Christ
 */
public class BasicAppSkeleton
{
	/**
	* Add new custom options to the list of options. These are in addition to the hard-coded options,
	* which means mostly the <tt>--help</tt> option.
	* <p>
	* This method will typically be overwritten by your specialization.
	* Your code will have to add specializations of the LongOpt class to the end of the options Vector.
	* <p>
	* Example:<p>
	* <code>options.addElement(new NumericalOpt(&quot;minimum&quot;, &quot;The minimum value&quot;, 3, true, 0, 5));<code>
	* <p>
	* @param options the options, as a Vector of LongOpt objects
	* @see com.syncbuilder.util.getopt.LongOpt
	* @see com.syncbuilder.util.getopt.NumericalOpt
	* @see com.syncbuilder.util.getopt.StringOpt
	* @see com.syncbuilder.util.getopt.FlagOpt
	*/
	public void addCustomOptions()	
	{
	}
	

	/**
	* The custom-code for your application goes here. It will be invoked by the skeleton after the required 
	* options have been asked from the user.
	*/
	public void appMain()
		throws Exception
	{
	}


	/**
	* Return a String which contains information about the program.
	* <p>
	* You will almost always overwrite this method in order to return
	* information about your program.
	*/
	public String getProgramInfo()
	{
		return "Welcome!";
	}
		
	
	/**
	* Invoke this method to run the application. It will eventually call your
	* appMain() method, which should contain the main portion of the program.
	* <p>
	* runApp handles all the command-line parsing for you. It will also display
	* help-texts if appropriate. It will only call your code if all the parameters
	* are valid.
	*/
	public void runApp(String[] args)
	{
		System.err.println();
		System.err.println(getProgramInfo());
		
		options = new Vector();
		FlagOpt helpopt = new FlagOpt("help", "Display usage information");
		helpopt.setSecret(true);
		options.addElement(helpopt);

		FlagOpt guiopt = new FlagOpt("gui", "Activate GUI interface");
		guiopt.setSecret(true);
		options.addElement(guiopt);
		
		addFixedOptions();
		addCustomOptions();

		Index xtra_args = new Index();
		_parseArguments(args, xtra_args);
		extra_args = xtra_args.getValue();

		if (_arguments == null)
		{
			_displayUsage();
			System.exit(0);
		}


		if (_arguments.containsKey("help"))
		{
			_displayUsage();
			System.exit(0);
		}

		if (_arguments.containsKey("gui"))
		{
			_has_gui = true;
		}
		
		ConsoleWindow cw = null;
		if (_has_gui)
		{
			OptionsWindow ow = new OptionsWindow(getProgramInfo(), options, _arguments);
			ow.run();
			ow.dispose();
			
			if (ow.isCancelled())
			{
				System.exit(0);
			}
			
			cw = new ConsoleWindow(getProgramInfo());
			System.setOut(cw.getOutStream());
			System.setErr(cw.getErrStream());
		}
		else
		{
			_inquireParameters();
		}

		try
		{
			appMain();
		}
		catch(Exception e)
		{
			displayMessage("Execution failed:\n" + e.getMessage(), "abort" );
		}

		if (_has_gui)
		{
			cw.waitForClose();
		}

		System.exit(0);
	}


	/**
	* Get the numerical value of an option
	*/
	public int getNumericalOpt(String name)
	{
		return Integer.parseInt((String)(_arguments.get(name)));
	}
	
	
	/**
	* Get the String value of an option
	*/
	public String getStringOpt(String name)
	{
		return (String)(_arguments.get(name));
	}
	
	
	/**
	* Get the boolean value of an option
	*/
	public boolean getFlagOpt(String name)
	{
		String flag = (String)(_arguments.get(name));
		
		if (flag == null)
			return false;
			
		if (flag.equals("true"))
			return true;
		else			
			return false;
	}
	

	/**
	* Use this method to display messages that require the immediate attention of the user.
	* This method will block until the user has acknowledged the message.
	*
	* @param msg the text of the message
	* @param action a description of what happens when the user acknowledges the message (e.g. continue).
	*/
	public void displayMessage(String msg, String action)
	{
		if (!_has_gui)
		{
			System.err.println();
			System.err.println(msg);
			System.err.println("Press <RETURN> to " + action + "...");

			try
			{
				com.syncbuilder.util.Util.readLine();
			} 
			catch(java.io.IOException ioe)
			{
			}
		}
		else
		{
			MessageBox.showMessage(msg, action);
		}
	}
	
	
	/**
	* Parse the arguments from the command line, and fill them into a Hashtable.
	* <br>You will typically not over-ride this method.
	*
	* @param args the args-array contains the command-line parameters
	* @param extra_args if this is <code>!= null</code> it will be filled with the index of the first extra-argument after the last command-line parameter
	*/
	protected void _parseArguments(String[] args, Index extra_args)
	{
		LongOpt[] longopts = new LongOpt[options.size()];
		
		options.copyInto(longopts);

		Getopt g = new Getopt("(ignored)", args, "W;", longopts, true);
		g.setOpterr(false); // We'll do our own error handling

		Hashtable parameters = new Hashtable();


		int c;
	 	String arg;
	 	boolean failure = false;

	 	while (((c = g.getopt()) != -1) && !failure)
		{
	   		switch (c)
    		{
        		case 2:
        			arg = g.getOptarg();
					if (arg == null)
					{
						arg = Boolean.TRUE.toString();  // This can only be a set flag
					}

					LongOpt opt = longopts[g.getLongind()];
					
					if (!opt.validateArgument(arg))
					{
						System.err.println("The parameter you passed to --" + opt.getName() + " is invalid.");
						failure = true;
						break;
					}
					
					parameters.put(opt.getName(), arg);
        		break;
        		  //
        		case 'W':
        			System.err.println("You tried a -W with an incorrect long option name.");
					failure = true;
        		break;
        		  //
        		case ':':
            		System.err.println("Option " + (char)g.getOptopt() + " needs an argument.");
					failure = true;
            	break;
            		//
        		case '?':
            		System.err.println("The option '" + args[g.getOptind() - 1] + "' is not valid");
					
					failure = true;
            	break;
            		//
				default:
        		break;
    		 }
		}

		if (failure)
		{
			_arguments = null;
			return;
		}

		if (extra_args != null)
		{
			extra_args.setValue(g.getOptind());
		}

		_arguments = parameters;
		return;
	}


	/**
	* Determine the missing parameters, either through the default-values, or by asking the user.
	*/
	protected void _inquireParameters()
	{
		boolean missing_printed = false;

		for (int i = 0; i < options.size(); i++)
		{
			LongOpt opt = (LongOpt)(options.elementAt(i));
			
			if (!_arguments.containsKey(opt.getName()))
			{
				while(true)
				{
					String value = (opt.getAutoDefault()) ? opt.getDefault() : null;
					
					if ( value == null)
					{
						if (!missing_printed)
						{
							missing_printed = true;
							System.err.println("\nPlease enter the following missing parameters:");
						}

						System.err.print("  " + Util.justifyLeft(opt.getDescription() + ": ", 50));

						try
						{
							value = Util.readLine().trim();
							
							if (value.length() == 0)
							{
								value = opt.getDefault();
								
								if (value == null)
								{
									value = "";
								}
							}
						}
						catch(IOException ioe)
						{
							System.exit(0);
						}
					}

					if (opt.validateArgument(value) == true)
					{
						_arguments.put(opt.getName(), value);
						break;
					}

				}
			}
		}
	}


	protected void _displayUsage()
	{
		System.err.println("Usage:");

		for (int i = 0; i < options.size(); i++)
		{
			LongOpt opt = (LongOpt)(options.elementAt(i));
			System.err.println("    --" + Util.justifyLeft(opt.getName(), 15) + ": " + opt.getDescription());
		}
	}


	/**
	* This one will be invoked by runApp before addCustomOptions(). It does the same thing.
	* This method should be used by base classes for certain types of command-line apps, 
	* whereas addCustomOptions() is meant for use by the final application.
	*
	* @see #addCustomOptions
	*/
	protected void addFixedOptions()
	{
	}


	protected Vector options;
	protected Hashtable _arguments;
	public int extra_args;
	protected boolean _has_gui = false;
}


class OptionsWindow 
	extends Frame
	implements ActionListener
{
    public OptionsWindow(String info, Vector options, Hashtable args)
	{
		_options = options;
		_args = args;

		component = new Component[options.size()];

		setLayout(new BorderLayout());

		Panel info_panel = new Panel();
		info_panel.setLayout(new BorderLayout());
		Label info_label = new Label(info);
		info_label.setFont(new Font("SansSerif", Font.BOLD, 12));
		info_panel.add("West", info_label);

		Panel button_panel = new Panel();
		button_panel.setLayout(new FlowLayout());
		ok_button = new Button("OK");
		cancel_button = new Button("Cancel");
		button_panel.add(ok_button);
		button_panel.add(cancel_button);
		ok_button.addActionListener(this);
		cancel_button.addActionListener(this);


		Panel opt_panel = new Panel();
        opt_panel.setLayout(new GridLayout(0,2));

		for (int i = 0; i < options.size(); i++)
		{
			LongOpt opt = ((LongOpt)(options.elementAt(i)));

			if (!opt.isSecret())
			{
				opt_panel.add(new Label(opt.getPureDescription()));

				String defaultval;
				
				if (args.containsKey(opt.getName()))
				{
					defaultval = ((String)(args.get(opt.getName())));
				}
				else
				{
					defaultval = opt.getDefault();
				}

				if (opt instanceof FlagOpt)
				{
					Checkbox cb = new Checkbox();
					if (defaultval != null)
					{
						if (defaultval.equals(Boolean.TRUE.toString()))
						{
							cb.setState(true);
						}
					}

					component[i] = cb;
					opt_panel.add(cb);
				}
				else
				{
					TextField tf = new TextField(20);
					if (defaultval != null)
					{
						tf.setText(defaultval);
					}
					
					component[i] = tf;
					opt_panel.add(tf);
				}
			}
		}

		add("North", info_panel);
		add("South", button_panel);
		add("Center", opt_panel);

		StringTokenizer tok = new StringTokenizer(info);
		setTitle(tok.nextToken());
		pack();
		show();
    }


	public synchronized void run()
	{
		while(!_finished)
		{
			try
			{
				wait();
			}
			catch(InterruptedException ie)
			{
			}
			
//			System.err.println("Run cycle");
		}
	}


	public boolean isCancelled()
	{
		return _cancelled;
	}

	public synchronized void actionPerformed(ActionEvent evt)
	{
		Object source = evt.getSource();
		
		if (source == ok_button)
		{
			boolean valid = true;

			for (int i = 0; i < _options.size(); i++)
			{
				LongOpt opt = ((LongOpt)(_options.elementAt(i)));
				
				if (!opt.isSecret())
				{
					String argument = null;
					if (opt instanceof FlagOpt)
					{
						argument = new Boolean(((Checkbox)(component[i])).getState()).toString();
					}
					else
					{
						argument = ((TextField)(component[i])).getText().trim();
					}

					if (!opt.validateArgument(argument))
					{
						component[i].requestFocus();
						valid = false;
						break;
					}
					else
					{
						_args.put(opt.getName(), argument);						
					}
				}
			}
			
			_finished = valid;
		}
		else if (source == cancel_button)
		{
			_finished = true;
			_cancelled = true;
		}

		notifyAll();
	}


	private boolean _finished = false;
	private boolean _cancelled = false;
	private Button ok_button;
	private Button cancel_button;

	private Vector _options;
	private Hashtable _args;

	private Component[] component;
}



class ConsoleWindow
	extends Frame
	implements ActionListener
{
	public ConsoleWindow(String info)
	{
		setLayout(new BorderLayout());

		_outstream = new TextAreaPrintStream(20, 80);		
		_errstream = new TextAreaPrintStream(10, 80);		

		Panel main_panel = new Panel();
		main_panel.setLayout(new BorderLayout());
		main_panel.add("North", _errstream.getTextArea());
		main_panel.add("Center", _outstream.getTextArea());

		Button close_button = new Button("Close");
		close_button.addActionListener(this);
		Panel close_panel = new Panel();
		close_panel.setLayout(new FlowLayout());
		close_panel.add(close_button);

		add("Center", main_panel);
		add("South", close_panel);
		
		StringTokenizer tok = new StringTokenizer(info);
		setTitle(tok.nextToken() + " - Console Window");
		pack();
		show();
	}


	public synchronized void waitForClose()
	{
		while(!_closed)
		{
			try
			{
				wait();
			}
			catch(InterruptedException ie)
			{
			}
		}
	}


	public synchronized void actionPerformed(ActionEvent evt)
	{
		dispose();
		_closed = true;
		notifyAll();
	}


	public PrintStream getOutStream()
	{
		return _outstream;
	}

	public PrintStream getErrStream()
	{
		return _errstream;
	}

	private TextAreaPrintStream _outstream;
	private TextAreaPrintStream _errstream;

	private boolean _closed = false;
}


class TextAreaPrintStream
	extends java.io.PrintStream
{
	public TextAreaPrintStream(int rows, int columns)
	{
		super(null);
		_area = new TextArea(rows, columns);
		_area.setEditable(false);
		_area.setFont(new Font("Monospaced", Font.PLAIN, 12));
		_line = new StringBuffer();
	}

	public void print(String text)
	{
		_line.append(text);
	}

	public void println(String text)
	{
		_line.append(text + "\n");
		flush();
	}

	public void println(Object obj)
	{
		println(obj.toString());
	}

	public void flush()
	{
		_area.append(_line.toString());
		_area.invalidate();
		_line = new StringBuffer();
	}


	public TextArea getTextArea()
	{
		return _area;
	}
	
	private TextArea _area;
	private StringBuffer _line;
}


class MessageBox
	extends Frame
	implements ActionListener
{
	public MessageBox(String msg, String action)
	{
		setLayout(new BorderLayout());

		TextArea area = new TextArea(msg, 5, 50, TextArea.SCROLLBARS_NONE);
		area.setEditable(false);

		Panel main_panel = new Panel();
		main_panel.setLayout(new BorderLayout());
		main_panel.add("Center", area);

		Button close_button = new Button(action);
		close_button.addActionListener(this);
		Panel close_panel = new Panel();
		close_panel.setLayout(new FlowLayout());
		close_panel.add(close_button);

		add("Center", main_panel);
		add("South", close_panel);
		
		setTitle("Attention");
		pack();
		show();
	}


	public static void showMessage(String message, String action)
	{
		MessageBox mb = new MessageBox(message, action);
		
		mb.waitForClose();
		mb.dispose();
	}	


	public synchronized void waitForClose()
	{
		while(!_closed)
		{
			try
			{
				wait();
			}
			catch(InterruptedException ie)
			{
			}
		}
	}



	public synchronized void actionPerformed(ActionEvent evt)
	{
		dispose();
		_closed = true;
		notifyAll();
	}

	private boolean _closed = false;
}
