/* com/syncbuilder/device/SysInfo.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: SysInfo.java,v 1.7 1999/01/24 12:31:09 christ Exp $ */

package com.syncbuilder.device;

/**
* This class stores fundamental information about the Palm device.
*
* @author Tilo Christ
*/
public class SysInfo 
{
	/**
	* Get the major version number of the Palm device's system ROM. Upgrade patches are taken into
	* account by this method.
	*/
	public int getRomVersionMajor()
	{
		return SystemFeature.getRomVersionMajor(_rom_sw_version);
	}

	/**
	* Get the minor version number of the Palm device's system ROM. Upgrade patches are taken into
	* account by this method.
	*/
	public int getRomVersionMinor()
	{
		return SystemFeature.getRomVersionMinor(_rom_sw_version);
	}

	/**
	* Get the version number of the Palm device's system ROM formatted as a displayable string (e.g. 2.05).
	* Upgrade patches are taken into account by this method.
	*/
	public String getRomVersionString()
	{
		return SystemFeature.getRomVersionString(_rom_sw_version);
	}


	/**
	* Get the 'LocaleID'.
	* Currently all devices will return 0x00010000 here.
	*/
	public int getLocaleID()
	{
		return _locale_id;
	}

	/**
	* Get the 'ProductID'.
	* Currently all devices will return 0x00010000 here.
	*/
	public int getProductID()
	{
		return _prod_id;
	}


	/**
	* Devices which feature DL 1.2+ (those with Palm OS 3.0+) will return <code>true</code> to
	* signal that extra information is available.
	*
	* @see #getDLPVersion
	* @see #getCompatibilityVersion
	* @see #getMaximumRecordSize
	*/
	public boolean isSysInfoVer()
	{
		return _sys_info_ver;
	}

	/**
	* Get the version of the DL protocol used by this device. This version number can be compared with
	* the DL_xxx constants from the Dlp class.
	*
	* This information is only valid if <code>isSysInfoVer()</code> returned <code>true</code>!
	*
	* @see #isSysInfoVer
	*/
	public int getDLPVersion()
	{
		return _dlp_version;
	}


	/**
	* The exact meaning of this number is unknown.
	*
	* This information is only valid if <code>isSysInfoVer()</code> returned <code>true</code>!
	*
	* @see #isSysInfoVer
	*/
	public int getCompatibilityVersion()
	{
		return _compatibility_version;
	}
	

	/**
	* Get the maximum size of a record that may be written to the device.
	* 0xFFFFFFFF means the size is only limited by the available memory.
	*
	* This information is only valid if <code>isSysInfoVer()</code> returned <code>true</code>!
	*
	* @see #isSysInfoVer
	*/
	public int getMaximumRecordSize()
	{
		return _maximum_record_size;
	}


	public String toString() 
	{
		return "<com.syncbuilder.device.SysInfo: ROM version "+ getRomVersionString() +", Locale ID 0x"+ Integer.toHexString(_locale_id) +
			", Product ID 0x" + Integer.toHexString(_prod_id) + 
			(
				(_sys_info_ver) ? 
			 		(",\nDLP version " + (_dlp_version >> 8) + "." + (_dlp_version & 0xFF) + ", Compatibility version " + (_compatibility_version >> 16) + "." + (_compatibility_version & 0xFFFF) + ", Maximum Record Size " + _maximum_record_size) 
			 	: 
					""
			)
			+ ">";
	}

	/**
	* <b>This constructor shall not be invoked by anybody, except DlpImpl!</b>
	*/
	public SysInfo(int rom_sw_version, int locale_id, int prod_id)
	{
		this._rom_sw_version = rom_sw_version;
		this._locale_id = locale_id;
		this._prod_id = prod_id;

		_sys_info_ver = false;
	}


	/**
	* <b>This constructor shall not be invoked by anybody, except DlpImpl!</b>
	*/
	public SysInfo(int rom_sw_version, int locale_id, int prod_id, int dlp_version, int comp_version, int max_rec_size)
	{
		this._rom_sw_version = rom_sw_version;
		this._locale_id = locale_id;
		this._prod_id = prod_id;

		this._dlp_version = (dlp_version >> 8) + (dlp_version & 0xff);  // Make compatible with Dlp.DL_xxx constants
		this._compatibility_version = comp_version;
		this._maximum_record_size = max_rec_size;

		_sys_info_ver = true;
	}

	private int _rom_sw_version;
	private int _locale_id;   // Currently all devices have 0x00010000 here
	private int _prod_id; // Currently all devices have 0x00010000 here

	private int _dlp_version;
	private int _compatibility_version; // Meaning unknown
	private int _maximum_record_size;

	private boolean _sys_info_ver;
}
