/* com/syncbuilder/internal/sync/DlpImpl.java:
*
* Copyright (C) 1998, 1999, Tilo Christ
* Copyright (C) 1998, Robert Crawford
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: DlpImpl.java,v 1.15 1999/02/21 13:39:56 christ Exp $ */

package com.syncbuilder.internal.sync;

import com.syncbuilder.internal.core.constants;

import com.syncbuilder.sync.DlpException;
import com.syncbuilder.sync.RPCParam;
import com.syncbuilder.storage.*;
import com.syncbuilder.device.*;

import com.syncbuilder.util.Index;
import com.syncbuilder.util.PackUtils;
import com.syncbuilder.util.Util;

/**
* This class implements the DLP in pure Java. It is here to
* avoid cluttering of the <code>com.syncbuilder.sync.Dlp</code> class, which 
* acts as a front-end to this class.<br>
* <b>You are not supposed to ever call any methods in this class directly!</b>
* <p>
* If you wish to learn more about the DLP protocol you should consult the file
* <tt>System/DLCommon.h</tt> from the Palm OS SDK. You may also wish to consult
* the document found at <a href="http://www.cdpubs.com/hhsys/archives/53/08flynn.pdf">
* http://www.cdpubs.com/hhsys/archives/53/08flynn.pdf</a>
* <p>
* The method names in this class are adapted from DLCommon.h, with the dlp-prefix stripped
* away.
*
* @see com.syncbuilder.sync.Dlp
* @author Tilo Christ, Robert Crawford
*/
public final class DlpImpl
{
//============================================== Database ===================================================
	/**
	* Open a database on the Palm device through its cardnumber and its name, using the specified mode.
	*
	* @return the handle of the database
	*/
	public int openDB(String dbname, int mode, int cardno)
		throws DlpException
	{
		byte[] dlp_buf = new byte[dbname.length() + 3];
		dlp_buf[0] = (byte)cardno;
		dlp_buf[1] = (byte)mode;
		PackUtils.setZeroTermString(dlp_buf, 2, dbname); 

		byte[] result = _dlp_exec(0x17, FIRSTARGID, dlp_buf);
  
	    _expect(result, 1);
  
	    int handle = result[0] & 0xff;

		return handle;
	}

	/**
	* Close a single open database.
	*/
	public void closeDB(int handle)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];
		dlp_buf[0] = (byte)handle;
		
		_dlp_exec(0x19, FIRSTARGID, dlp_buf);
	}


	public int createDB(String dbname, Char4 creator, Char4 type, int cardno, int flags, int version)
		throws DlpException
	{
		byte[] dlp_buf = new byte[dbname.length() + 15];

		PackUtils.setLong(dlp_buf, 0, creator.getInt());
		PackUtils.setLong(dlp_buf, 4, type.getInt());
		dlp_buf[8] = (byte)cardno;
		dlp_buf[9] = 0;
		PackUtils.setShort(dlp_buf, 10, flags);
		PackUtils.setShort(dlp_buf, 12, version);
		PackUtils.setZeroTermString(dlp_buf, 14, dbname);

		byte[] result = _dlp_exec(0x18, FIRSTARGID, dlp_buf);

		_expect(result, 1);

		return (result[0] & 0xff);
	}


	public void deleteDB(String dbname, int card)
		throws DlpException
	{
		byte[] dlp_buf = new byte[dbname.length() + 3];
		
		dlp_buf[0] = (byte)card;
		dlp_buf[1] = 0;
		
		PackUtils.setZeroTermString(dlp_buf, 2, dbname);

		try
		{
			_dlp_exec(0x1A, FIRSTARGID, dlp_buf);
		} 
		catch (DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
			{
				return;  // Database not found will be silently ignored
			}
			else
			{
				throw de;
			}
		}
	}

	
	public void cleanUpDatabase(int handle)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];

		dlp_buf[0] = (byte)handle;
		
		_dlp_exec(0x26, FIRSTARGID, dlp_buf);
	}


	public void resetSyncFlags(int handle)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];

		dlp_buf[0] = (byte)handle;
		
		_dlp_exec(0x27, FIRSTARGID, dlp_buf);
	}
		

	/**
	* Get meta-information about a <i>single</i> database on the Palm device.
	* <br>(the method name suggests that this method can read a list of database information, but
	* protocol versions below DL 1.2 do not support that).
	*
	* @return the meta-information about the database, or <code>null</code> if the database does not exist
	*/
	public DBInfo readDBList(int index, int cardno, int flags)
		throws DlpException
	{
		byte[] dlp_buf = new byte[4];
		dlp_buf[0] = (byte)flags;
		dlp_buf[1] = (byte)cardno;
		PackUtils.setShort(dlp_buf, 2, index);
		
		byte[] result = null;

		try
		{
			result = _dlp_exec(0x16, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound database is indicated by null instead of an exception
			else
				throw de;
		}
		
		_expect(result, 48);
		
//		boolean more = (result[2] != 0); 


		return _extractDBInfo(result, 4, cardno);
	}


	/**
	* Retrieve the number of records in the specified database.
	*/
	public int readOpenDBInfo( int handle )
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];
			
		dlp_buf[0] = (byte)handle;
	    byte[] result = _dlp_exec(0x2B, FIRSTARGID, dlp_buf);

	    _expect(result, 2);
  
    	return PackUtils.getShort(result, 0);
	}


	/**
	* Retrieve a list of all the records in the specified database.
	* The records will be identified through their UID.
	*/
	public RecordID[] readRecordIDList(int handle, boolean sort, int start, int max)
		throws DlpException
	{
		byte[] dlp_buf = new byte[6];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)(sort ? 0x80 : 0x00);  // 0x80 = dlpReadRecordIDListFlagSortDB
		PackUtils.setShort(dlp_buf, 2, start);
		PackUtils.setShort(dlp_buf, 4, max);
		
		byte[] result = _dlp_exec(0x31, FIRSTARGID, dlp_buf);
		
		_expect(result, 2);
		
		int records = PackUtils.getShort(result, 0);
		
		RecordID[] rec_ids = new RecordID[records];
		
		for (int i = 0; i < records; i++)
		{
			rec_ids[i] = new RecordID(PackUtils.getLong(result, 2 + 4*i));
		}
		
		return rec_ids;
	}


	public void moveCategory(int handle, int from, int to)
		throws DlpException
	{
		byte[] dlp_buf = new byte[4];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)from;
		dlp_buf[2] = (byte)to;
		dlp_buf[3] = (byte)0;
		
		_dlp_exec(0x2C, FIRSTARGID, dlp_buf);
	}


	public DBInfo findDBByOpenHandle(int handle)
		throws DlpException
	{
		// This DLP-function is only implemented in DL 1.2+. 
		// It cannot be emulated on lower versions.
		if (_version >= com.syncbuilder.sync.Dlp.DL_1_2)
		{
			byte[] dlp_buf = new byte[2];

			dlp_buf[0] = (byte)0x80; // dlpFindDBOptFlagGetAttributes
			dlp_buf[1] = (byte)handle;

			byte[] result = _dlp_exec(0x39, FIRSTARGID + 1, dlp_buf);

			_expect(result, 54);

			int cardno = result[0] & 0xff;

			return _extractDBInfo(result, 10, cardno);
		}
		else 
		{
			throw new DlpException(13);  // Operation not supported
		}
	}


	public DBInfo findDBByName(int card, String name)
		throws DlpException
	{
		// This DLP-function is only implemented in DL 1.2+. 
		// We can emulate it on lower versions, but the costs are high!
		if (_version >= com.syncbuilder.sync.Dlp.DL_1_2)
		{
			byte[] dlp_buf = new byte[2 + name.length() + 1];
			
			dlp_buf[0] = (byte)0x80; // dlpFindDBOptFlagGetAttributes
			dlp_buf[1] = (byte)card;
			
			PackUtils.setZeroTermString(dlp_buf, 2, null);
			
			byte[] result = _dlp_exec(0x39, FIRSTARGID, dlp_buf);
			
			_expect(result, 54);

			return _extractDBInfo(result, 10, card);
		}
		else
		{
			// Search only RAM databases first
			int i = 0;
			while(true)
			{
				DBInfo dbinfo = readDBList(i, card, com.syncbuilder.internal.core.constants.dlpDBListRAM);

				if (dbinfo == null) 
				{
					break;
				}
				else
				{
					if (dbinfo.getName().equals(name))
					{
						return dbinfo;
					}
					
					i = dbinfo.getIndex() + 1;
				}
			}


			// ... then search only ROM databases.
			i = 0;
			while(true)
			{
				DBInfo dbinfo = readDBList(i, card, com.syncbuilder.internal.core.constants.dlpDBListROM);

				if (dbinfo == null) 
				{
					break;
				}
				else
				{
					if (dbinfo.getName().equals(name))
					{
						return dbinfo;
					}
					
					i = dbinfo.getIndex() + 1;
				}
			}

			// Nothing could be found
			return null;
		}		
	}

	
//===================================================== Record ===============================================

	/**
	* Retrieve a Record from the database through its index. 
	*
	* @return <code>null</code> if the record could not be found, a specialization of <code>com.syncbuilder.storage.Record</code> otherwise.
	*/
	public Record readRecordByIndex(int handle, int index, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];

		dlp_buf[0] = (byte)handle;
  		dlp_buf[1] = (byte)0x00;
  		PackUtils.setShort(dlp_buf, 2, index);
  		PackUtils.setShort(dlp_buf, 4, 0); // Offset into record 
  		PackUtils.setShort(dlp_buf, 6, DLP_BUF_SIZE); // length to return 

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x20, FIRSTARGID + 1, dlp_buf);  // FIRSTARGID + 1 = dlpReadRecordIndexArgID
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		return _unpackRecord(result, dbimpl);
	}


	/**
	* Retrieve a Record from the database through its UID. 
	*
	* @return <code>null</code> if the record could not be found, a specialization of <code>com.syncbuilder.storage.Record</code> otherwise.
	*/
	public Record readRecordByID(int handle, RecordID req_id, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];

		dlp_buf[0] = (byte)handle;
  		dlp_buf[1] = (byte)0x00;
  		PackUtils.setLong(dlp_buf, 2, req_id.getID());
  		PackUtils.setShort(dlp_buf, 6, 0); // Offset into record 
  		PackUtils.setShort(dlp_buf, 8, DLP_BUF_SIZE); // length to return 

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x20, FIRSTARGID, dlp_buf);  // FIRSTARGID = dlpReadRecordIdArgID
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		return _unpackRecord(result, dbimpl);
	}


	public Record readNextModifiedRec(int handle, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];

		dlp_buf[0] = (byte)handle;

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x1F, FIRSTARGID, dlp_buf);  
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		return _unpackRecord(result, dbimpl);
	}


	public Record readNextRecInCategory(int handle, int category, DatabaseImpl dbimpl)
		throws DlpException
	{
		// FIXME: Emulate for PalmOS 1.0

		byte[] dlp_buf = new byte[2];

		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)category;

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x32, FIRSTARGID, dlp_buf);  
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		return _unpackRecord(result, dbimpl);
	}


	public Record readNextModifiedRecInCategory(int handle, int category, DatabaseImpl dbimpl)
		throws DlpException
	{
		// FIXME: Emulate for PalmOS 1.0

		byte[] dlp_buf = new byte[2];

		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)category;

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x33, FIRSTARGID, dlp_buf);  
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		return _unpackRecord(result, dbimpl);
	}

	/** 
	* Resets the &quot;next record&quot; index to the beginning 
	*/
	public void resetRecordIndex(int handle)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];
		dlp_buf[0] = (byte)handle;
		
		_dlp_exec(0x30, FIRSTARGID, dlp_buf);				
	}


	/**
	* Write the specified record to the specified database.
	* If the Record had a RecordID of 0 before, it will be
	* assigned a valid RecordID afterwards.
	*/
	public void writeRecord(int handle, Record record)
		throws DlpException
	{
		byte[] content = record.pack();

		byte[] dlp_buf = new byte[content.length + 8];

	    dlp_buf[0] = (byte)handle;
	  	dlp_buf[1] = (byte)0;
		PackUtils.setLong(dlp_buf, 2, record.getID().getID());
	  	dlp_buf[6] = (byte)(record.getFlags());
	  	dlp_buf[7] = (byte)(record.getCategory());

	  	System.arraycopy(content, 0, dlp_buf, 8, content.length);

	  	byte[] result = _dlp_exec(0x21, FIRSTARGID, dlp_buf);

	  	_expect(result, 4);

      	record.getID().setID(PackUtils.getLong(result, 0)); /* New record ID */
	}


	/**
	* Delete records from a database. 
	* 
	* @param handle DB handle
	* @param flags 0x00 - delete rec with specified id, 0x80 - delete ALL recs, 0x40 - delete all recs with category id
	* @param id record id, or category id (when flag is 0x40). Ignored when flag is 0x80.
	*/
	public void deleteRecord(int handle, int flags, int id)
		throws DlpException
	{
		// FIXME: PalmOS 1.0 doesn't support the 0x40-flag!

		byte[] dlp_buf = new byte[6];

		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)flags;
		PackUtils.setLong(dlp_buf, 2, id);

		try
		{
			_dlp_exec(0x22, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return;    // unfound record is silently ignored
			else
				throw de;
		}
	}

	
//===================================================== Resource ==========================================

	public Resource readResourceByIndex(int handle, int index, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)0;
		PackUtils.setShort(dlp_buf, 2, index);
		PackUtils.setShort(dlp_buf, 4, 0);
		PackUtils.setShort(dlp_buf, 6, 0xFFFF);

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x23, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound resource is silently ignored
			else
				throw de;
		}

		return _unpackResource(result, dbimpl);
	}



	public Resource readResourceByType(int handle, Char4 type, int id, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[12];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)0;
		PackUtils.setLong(dlp_buf, 2, type.getInt());
		PackUtils.setShort(dlp_buf, 4, id);
		PackUtils.setShort(dlp_buf, 8, 0);
		PackUtils.setShort(dlp_buf, 10, 0xFFFF);

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x23, FIRSTARGID + 1, dlp_buf);  // FIRSTARGID + 1 = dlpReadResourceTypeArgID
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound resource is silently ignored
			else
				throw de;
		}

		return _unpackResource(result, dbimpl);
	}


	public void writeResource(int handle, Resource resource)
		throws DlpException
	{
		byte[] content = resource.pack();
		byte[] dlp_buf = new byte[content.length + 10];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)0;
		PackUtils.setLong(dlp_buf, 2, resource.getType().getInt());
		PackUtils.setShort(dlp_buf, 6, resource.getID());
		PackUtils.setShort(dlp_buf, 8, content.length);
		System.arraycopy(content, 0, dlp_buf, 10, content.length);
		
		_dlp_exec(0x24, FIRSTARGID, dlp_buf);
	}


	public void deleteResource(int handle, boolean all, Char4 type, int id)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];
		
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)(all ? 0x80 : 0x00); // 0x80 = dlpDeleteResFlagDeleteAll

		if (!all)
		{
			PackUtils.setLong(dlp_buf, 2, type.getInt());
			PackUtils.setShort(dlp_buf, 6, id);
		}
		
		_dlp_exec(0x25, FIRSTARGID, dlp_buf);
	}

//=======================  Preference ===========================================================

	public com.syncbuilder.storage.Pref readAppPreference(Char4 creator, int id, boolean backed_up, 
			DatabaseImpl dbimpl)
		throws DlpException, NoCreatorException
	{
		if (creator == null)
			throw new NoCreatorException();

// FIXME: Reimplement emulation for Palm OS 1.0

		byte[] dlp_buf = new byte[10];

		PackUtils.setLong(dlp_buf, 0, creator.getInt());
		PackUtils.setShort(dlp_buf, 4, id);
		PackUtils.setShort(dlp_buf, 6, 0xffff); // dlpReadAppPrefActualSize = 0xffff
    	dlp_buf[8] = (byte) (backed_up ? 0x80 : 0); // dlpAppPrefReqFlagBackedUp = 0x80
    	dlp_buf[9] = 0; 	// Reserved 

		byte[] result = null;

		try
		{
			result = _dlp_exec(0x34, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound preference is silently ignored
			else
				throw de;
		}

		_expect(result, 6);

    	int version = PackUtils.getShort(result, 0);

    	int total_size = PackUtils.getShort(result, 2);
    	int returned_size = PackUtils.getShort(result, 4);

		if (total_size != returned_size)
			throw new DlpException("DlpImpl.readAppPreference: Size mismatch!");

		byte[] content = new byte[total_size];

    	System.arraycopy(result, 6, content, 0, total_size);

    	return dbimpl.createPref(content, id, version, backed_up);
	}


	public void writeAppPreference(com.syncbuilder.storage.Pref pref)
		throws DlpException
	{
		byte[] content = pref.pack();
	
		byte[] dlp_buf = new byte[content.length + 12];

		PackUtils.setLong(dlp_buf, 0, pref.creator.getInt());
		PackUtils.setShort(dlp_buf, 4, pref.id);
		PackUtils.setShort(dlp_buf, 6, pref.version);
		PackUtils.setShort(dlp_buf, 8, content.length);
		dlp_buf[10] = (byte)(pref.backed_up ? 0x80 : 0x00); // dlpAppPrefReqFlagBackedUp = 0x80
		dlp_buf[11] = (byte)0;
		System.arraycopy(content, 0, dlp_buf, 12, content.length);
		
		_dlp_exec(0x35, FIRSTARGID, dlp_buf);
	}


//=======================  AppBlock & SortBlock ==================================================


	public com.syncbuilder.storage.AppBlock readAppBlock(int handle, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[6];
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = 0;
		PackUtils.setShort(dlp_buf, 2, 0);
		PackUtils.setShort(dlp_buf, 4, 0xffff);

		byte[] result = null;

		try
		{
		    result = _dlp_exec(0x1B, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound AppBlock is silently ignored
			else
				throw de;
		}

	  	_expect(result, 2);

		byte[] content = new byte[result.length - 2];

    	System.arraycopy(result, 2, content, 0, content.length);

		return dbimpl.createAppBlock(content);
	}


	public void writeAppBlock(int handle, com.syncbuilder.storage.AppBlock appblock)
		throws DlpException
	{
		byte[] block = appblock.pack();

		byte[] dlp_buf = new byte[block.length + 4];
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)0;
		PackUtils.setShort(dlp_buf, 2, block.length);
		
		System.arraycopy(block, 0, dlp_buf, 4, block.length);
		
		_dlp_exec(0x1C, FIRSTARGID, dlp_buf);
	}


	public com.syncbuilder.storage.SortBlock readSortBlock(int handle, DatabaseImpl dbimpl)
		throws DlpException
	{
		byte[] dlp_buf = new byte[6];
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = 0;
		PackUtils.setShort(dlp_buf, 2, 0);
		PackUtils.setShort(dlp_buf, 4, 0xffff);

		byte[] result = null;

		try
		{
		    result = _dlp_exec(0x1D, FIRSTARGID, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.getCode() == DlpException.NOT_FOUND)
				return null;    // unfound SortBlock is silently ignored
			else
				throw de;
		}

	  	_expect(result, 2);

		byte[] content = new byte[result.length - 2];

    	System.arraycopy(result, 2, content, 0, content.length);

		return dbimpl.createSortBlock(content);
	}


	public void writeSortBlock(int handle, com.syncbuilder.storage.SortBlock sortblock)
		throws DlpException
	{
		byte[] block = sortblock.pack();

		byte[] dlp_buf = new byte[block.length + 4];
		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte)0;
		PackUtils.setShort(dlp_buf, 2, block.length);
		
		System.arraycopy(block, 0, dlp_buf, 4, block.length);
		
		_dlp_exec(0x1E, FIRSTARGID, dlp_buf);
	}

//======================================== Palm device ======================================================

	public int readFeature(Char4 creator, int number)
		throws DlpException, NoCreatorException
	{
		if (creator == null)
			throw new NoCreatorException();
	
		// FIXME: Emulate for Palm OS 1.0

		byte[] dlp_buf = new byte[6];
		
		PackUtils.setLong(dlp_buf, 0, creator.getInt());
		PackUtils.setShort(dlp_buf, 4, number);
		
		byte[] result = _dlp_exec(0x38, FIRSTARGID, dlp_buf);

		_expect(result, 4);
		
		return PackUtils.getLong(result, 0);
	}


	public SysInfo readSysInfo()
		throws DlpException
	{
		if (_version >= com.syncbuilder.sync.Dlp.DL_1_2)
			return readSysInfoVer();

		byte[] result = _dlp_exec(0x12, 0x00, null);
		
		_expect(result, 14);
		
		int rom_sw_version = PackUtils.getLong(result, 0);
		int localization_id = PackUtils.getLong(result, 4);
		// result[8] is padding
		int prod_id_size = result[9] & 0xff; // This should always be 4
		int prod_id = PackUtils.getLong(result, 10);
		
		return new SysInfo(rom_sw_version, localization_id, prod_id);
	}

	
	/**
	* Read extended system information that is only available under DL 1.2+
	*/
	public SysInfo readSysInfoVer()
		throws DlpException
	{
		byte[] dlp_buf = new byte[4];
		PackUtils.setShort(dlp_buf, 0, 1);  // We have come in peace. We speak DL 1.2 :-)
		PackUtils.setShort(dlp_buf, 2, 2);

		byte[][] result = _dlp_multi_exec(0x12, FIRSTARGID, dlp_buf);

		_expect(result[0], 14);
		_expect(result[1], 12);
		
		int rom_sw_version = PackUtils.getLong(result[0], 0);
		int localization_id = PackUtils.getLong(result[0], 4);
		// result[0][8] is padding
		int prod_id_size = result[0][9] & 0xff; // This should always be 4
		int prod_id = PackUtils.getLong(result[0], 10);

		int dlp_version = PackUtils.getLong(result[1], 0);
		int comp_version = PackUtils.getLong(result[1], 4);
		int max_record_size = PackUtils.getLong(result[1], 8);
		
		return new SysInfo(rom_sw_version, localization_id, prod_id, dlp_version, comp_version, max_record_size);
	}


	public java.util.Date getSysDateTime()
		throws DlpException
	{
		byte[] result = _dlp_exec(0x13, 0x00, null);
		
		_expect(result, 8);
		
		return _getDate(result, 0);
	}


	public void setSysDateTime(java.util.Date date)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];		

		_setDate(dlp_buf, 0, date);
		
		_dlp_exec(0x14, FIRSTARGID, dlp_buf);
	}
	

	/**
	* Read information about the HotSync user.
	*/	
	public UserInfo readUserInfo()
		throws DlpException
	{
    	byte[] result = _dlp_exec(0x10, 0x00, null);

    	_expect(result, 30);

    	int user_len = result[28] & 0xff;

		UserInfo uinfo = new UserInfo();

    	uinfo.userID = PackUtils.getLong(result, 0);
    	uinfo.viewerID = PackUtils.getLong(result, 4);
    	uinfo.lastSyncPC = PackUtils.getLong(result, 8);
    	uinfo.successfulSyncDate = _getDate(result, 12);
    	uinfo.lastSyncDate = _getDate(result, 20);

		int username_len = result[28] & 0xff;

		if (username_len == 0)
			return null;  // No user information has been established yet.
		
    	int password_len = result[29] & 0xff;
    	uinfo.username = PackUtils.getZeroTermString(result, 30, null);
		uinfo.password = new byte[password_len];
    	System.arraycopy(result, 30 + user_len, uinfo.password, 0, password_len);

    	return uinfo;
	}


	public void writeUserInfo(UserInfo uinfo)
		throws DlpException
	{
		int username_length = uinfo.username.length() + 1;
		byte[] dlp_buf = new byte[22 + username_length];
		
		PackUtils.setLong(dlp_buf, 0, uinfo.userID);
		PackUtils.setLong(dlp_buf, 4, uinfo.viewerID);
		PackUtils.setLong(dlp_buf, 8, uinfo.lastSyncPC);
		_setDate(dlp_buf, 12, uinfo.lastSyncDate);
		dlp_buf[20] = (byte)(0x80 | 0x40 | 0x20 | 0x10 | 0x08);
		/* Change everything
		dlpUserInfoModUserID		0x80	(changing the user id)
		dlpUserInfoModSyncPC		0x40	(changing the last sync PC)
		dlpUserInfoModSyncDate		0x20	(changing sync date)
		dlpUserInfoModName			0x10	(changing user name)
		dlpUserInfoModViewerID		0x08	(changing the viewer id)
		*/ 
				
		dlp_buf[21] = (byte)username_length;
		
		PackUtils.setZeroTermString(dlp_buf, 22, uinfo.username);

		_dlp_exec(0x11, FIRSTARGID, dlp_buf);
	}


	public NetSyncInfo readNetSyncInfo()
		throws DlpException
	{
		byte[] result = _dlp_exec(0x36, 0x00, null);
		
		_expect(result, 24);
		
		boolean lan_sync = (result[0] != 0);
		// result[0] - result[17] are reserved
		int sync_pc_name_size = PackUtils.getShort(result, 18);
		int sync_pc_address_size = PackUtils.getShort(result, 20);
		int sync_pc_mask_size = PackUtils.getShort(result, 22);

		// Empty Strings are marked by a null-byte, and not through a size of zero.
		String sync_pc_name = (sync_pc_name_size > 1) ? PackUtils.getZeroTermString(result, 24, null) : null;
		String sync_pc_address = (sync_pc_address_size > 1) ? PackUtils.getZeroTermString(result, 24+sync_pc_name_size, null) : null;
		String sync_pc_mask = (sync_pc_mask_size > 1) ? PackUtils.getZeroTermString(result, 24+sync_pc_name_size+sync_pc_address_size, null) : null;

		return new NetSyncInfo(lan_sync, sync_pc_name, sync_pc_address, sync_pc_mask);
	}


	public void writeNetSyncInfo(NetSyncInfo nsinfo)
		throws DlpException
	{
		int sync_pc_name_size = nsinfo.getHostName().length() + 1;
		int sync_pc_address_size = nsinfo.getHostAddress().length() + 1;
		int sync_pc_mask_size = nsinfo.getHostSubnetMask().length() + 1;

		byte[] dlp_buf = new byte[24 + sync_pc_name_size + sync_pc_address_size + sync_pc_mask_size];

		dlp_buf[0] = (byte)(0x80 | 0x40 | 0x20 | 0x10);
		/*
		dlpNetSyncInfoModLanSyncOn	0x80	(changing the "lan sync on" setting)
		dlpNetSyncInfoModSyncPCName	0x40	(changing the sync PC host name)
		dlpNetSyncInfoModSyncPCAddr	0x20	(changing the sync PC address)
		dlpNetSyncInfoModSyncPCMask	0x10	(changing the sync PC subnet mask)
		*/
		dlp_buf[1] = (byte)(nsinfo.getLANSync() ? 0xFF : 0x00);
		PackUtils.setLong(dlp_buf, 2, 0);
		PackUtils.setLong(dlp_buf, 6, 0);
		PackUtils.setLong(dlp_buf, 10, 0);
		PackUtils.setLong(dlp_buf, 14, 0);

		PackUtils.setShort(dlp_buf, 18, sync_pc_name_size);
		PackUtils.setShort(dlp_buf, 20, sync_pc_address_size);
		PackUtils.setShort(dlp_buf, 22, sync_pc_mask_size);

		int index = 24;
		index = PackUtils.setZeroTermString(dlp_buf, index, nsinfo.getHostName());		
		index = PackUtils.setZeroTermString(dlp_buf, index, nsinfo.getHostAddress());		
		index = PackUtils.setZeroTermString(dlp_buf, index, nsinfo.getHostSubnetMask());
		
		_dlp_exec(0x37, FIRSTARGID, dlp_buf);
	}
	
	
	/**
	* This method serves two purposes:
	* <ul>
	* <li>Switch the message on the Palm device from
	* 'Identifying User' to 'Synchronizing'.
	* <li>Probe for user cancel.
	* </ul>
	*
	* @exception com.syncbuilder.sync.CancelSyncException The user cancelled the synchronization procedure
	*/
	public void openConduit()
		throws com.syncbuilder.sync.CancelSyncException, DlpException
	{
		_dlp_exec(0x2E, 0x00, null);
	}

	
	public void addSyncLogEntry(String entry)
		throws DlpException
	{
		byte[] dlp_buf = new byte[entry.length() + 1];
		
		PackUtils.setZeroTermString(dlp_buf, 0, entry);
	
		_dlp_exec(0x2A, FIRSTARGID, dlp_buf);
	}


	/**
	* End the HotSync procedure
	*/
	public void endOfSync(int status)
		throws java.io.IOException, DlpException
	{
		byte[] dlp_buf = new byte[2];
		PackUtils.setShort(dlp_buf, 0, status);

		_dlp_exec(0x2F, FIRSTARGID, dlp_buf);
	}


	public StorageInfo readStorageInfo(int card)
		throws DlpException
	{
		byte[] dlp_buf = new byte[2];
		dlp_buf[0] = (byte)card;
		dlp_buf[1] = 0;
		
		byte[][] result = _dlp_multi_exec(0x15, FIRSTARGID, dlp_buf);
		
		_expect(result[0], 29);

		// * This info is global *
		// result[0][0] is card number of last card retrieved
		// boolean more = (result[0][1] != 0);
		// result[0][2] is unused
		// result[0][3] is count of returned structures

		// * This info is on a per-card basis, but we assume there will be only one card *
		// result[0][4] is length of current CardInfo
		// result[0][5] is number of card
		int version = PackUtils.getShort(result[0], 6);
		java.util.Date creation = _getDate(result[0], 8);
		int rom_size = PackUtils.getLong(result[0], 16);
		int ram_size = PackUtils.getLong(result[0], 20);
		int ram_free = PackUtils.getLong(result[0], 24);
		int name_length = result[0][28] & 0xff;						
		int manufacturer_length = result[0][29] & 0xff;

		String name = new String(result[0], 0, 30, name_length);
		String manufacturer = new String(result[0], 0, 30+name_length, manufacturer_length);

		CardInfo card_info = new CardInfo(card, version, creation, rom_size, ram_size, ram_free, name, manufacturer);

		if (result.length == 1)
		{  
			// DL 1.0
			StorageInfo storage_info = new StorageInfo();
			storage_info.addCardInfo(card_info);
			return storage_info;			
		}

		// DL 1.1+
		_expect(result[1], 20);
		
		int ram_db_count = PackUtils.getShort(result[1], 0);
		int rom_db_count = PackUtils.getShort(result[1], 2);
		// result[1][4] ... result[1][20] are reserved

		StorageInfo storage_info = new StorageInfo(ram_db_count, rom_db_count);
		storage_info.addCardInfo(card_info);
		return storage_info;			
	}


	public void resetSystem()
		throws DlpException
	{
		_dlp_exec(0x29, 0x00, null);
	}


	/**
	* Send an RPC for execution by the TouchDown RPC Executor (which is part of the Palm OS). 
	* This is a poorly documented, high-risk method of executing arbitrary Palm OS system-functions
	* through a DLP connection.
	*/
	public void processRPC(int trap, Index a0, Index d0, RPCParam[] params)
		throws DlpException
	{
		int param_length = 0;
	
		for (int i = 0; i < params.length; i++)
		{
			param_length += params[i].getSize();
		}
		
		byte[] rpc_buffer = new byte[16 + param_length + params.length * 2];
		
		rpc_buffer[0] = 0x2D;
		rpc_buffer[1] = 0x01;
		rpc_buffer[2] = 0x00;
		rpc_buffer[3] = 0x00;
		
		PackUtils.setShort(rpc_buffer, 4, trap);
		PackUtils.setLong(rpc_buffer, 6, d0.getValue());
		PackUtils.setLong(rpc_buffer, 10, a0.getValue());

		PackUtils.setShort(rpc_buffer, 14, params.length);
		
		int index = 16;
		for (int i = params.length - 1; i >= 0; i--)
		{
			rpc_buffer[index] = (byte)((params[i].isRef()) ? 1 : 0);
			rpc_buffer[index + 1] = (byte)(params[i].getSize());
			
			System.arraycopy(params[i].getBytes(), 0, rpc_buffer, index + 2, params[i].getSize());
			
			index += params[i].getSize() + 2;
		}

		System.err.println("RPC in:\n" + Util.hexPrint(rpc_buffer));
		
		byte[] rpc_result = new byte[rpc_buffer.length + 2];

		try
		{
			_out.write(rpc_buffer, 0, rpc_buffer.length);
			_in.read(rpc_result, 0, rpc_result.length);
		} 
		catch( java.io.IOException ioe )
		{
      		throw new DlpException(1);
		}

		System.err.println("RPC out:\n" + Util.hexPrint(rpc_result));

		int err = PackUtils.getShort(rpc_result, 2);
		
		if (err != 0)
		{
			throw new DlpException("RPC error: " + err);
		}

		d0.setValue(PackUtils.getLong(rpc_result, 8));
		a0.setValue(PackUtils.getLong(rpc_result, 12));

		index = 18;
		for (int i = params.length - 1; i >= 0; i--)
		{
			if (rpc_result[index] != 0)
			{
				System.arraycopy(rpc_result, index + 2, params[i].getBytes(), 0, params[i].getSize());
			}
					
			index += params[i].getSize() + 2;
		}
	}


	/**
	* This method is not part of the DL protocol. It returns the version number of
	* the protocol that this instance of DlpImpl is currently speaking.
	*/
	public int getVersion()
	{
		return _version;
	}


//============================================== Internal to DlpImpl =========================================

	/**
	* <b>!!!This constructor shall only be invoked by the Dlp class!!!</b>
	*/
	public DlpImpl( com.syncbuilder.sync.Socket socket, int version )
		throws java.io.IOException
	{
		this._version = version;

		_in  = socket.getInputStream();
		_out = socket.getOutputStream();
	}


	/**
	* <b>!!!This constructor shall only be invoked by the xxxxSocketImpl classes!!!</b>
	* The DlpImpl which gets instantiated here is used for probing of the DL version.
	*/
	public DlpImpl( java.io.InputStream is, java.io.OutputStream os)
		throws java.io.IOException
	{
		// Fake DL 1.2 to enable invokation of readSysInfoVer()
		_version = com.syncbuilder.sync.Dlp.DL_1_2;

		_in  = is;
		_out = os;
	}	


	/**
	* Execute a DLP-command and return a single result.
	*/	
	private synchronized byte[] _dlp_exec(int cmd, int argid, final byte[] msg)
		throws DlpException
	{
		byte[][] result = _dlp_multi_exec(cmd, argid, msg);
		
		if (result == null)
			return null;
		else
			return result[0];
	}


	/**
	* Execute a DLP-command and return the set of results.
	*/	
	private synchronized byte[][] _dlp_multi_exec(int cmd, int argid, final byte[] msg)
		throws DlpException
	{
		int i;

		byte[] exec_buf = new byte[DLP_BUF_SIZE];
	    exec_buf[0] = (byte)cmd;

	  	if ((msg != null) && (argid != 0) && (msg.length != 0))
	  	{
			// We do have an argument
    		System.arraycopy(msg, 0, exec_buf, 6, msg.length);
    		exec_buf[1] = (byte)1;
    		exec_buf[2] = (byte)(argid | 0x80); // 0x80 = dlpSmallArgFlag
	    	exec_buf[3] = (byte)0;
    		PackUtils.setShort(exec_buf, 4, msg.length);
    		i = msg.length + 6;
		} 
		else 
		{
			// No arguments
    		exec_buf[1] = (byte)0;
    		i = 2;
	  	}

		try
		{
			_out.write(exec_buf, 0, i);
		  	i = _in.read(exec_buf, 0, DLP_BUF_SIZE);
		} 
		catch( java.io.IOException ioe )
		{
      		throw new DlpException(1);
		}

	  	int err = PackUtils.getShort(exec_buf, 2);

	  	if (err != 0) 
	  	{
			if (err == 18)
				throw new com.syncbuilder.sync.CancelSyncException();
			else
	    		throw new DlpException(err);
	  	}

	  	if (exec_buf[0] != (byte)(cmd | 0x80)) 
		{ /* received wrong response */
      		throw new DlpException(1);
	  	}

		int argument_count = exec_buf[1] & 0xff;

	  	if (argument_count == 0) /* no return blocks */
    		return null; 

		byte[][] result = new byte[argument_count][];

		int index = 4;
		int len;

		// Process all the return blocks
		for( i = 0; i < argument_count; i++)
		{
			if ((exec_buf[index] & 0xC0) == 0xC0) 
			{ /* Long arg */
	  			len = PackUtils.getLong(exec_buf, index + 2);
				result[i] = new byte[len];			
  				System.arraycopy(exec_buf, index + 6, result[i], 0, len);
				
				index += (6 + len);
	  		} 
			else if ((exec_buf[index] & 0x80) != 0) 
    		{ /* Short arg */
	  			len = PackUtils.getShort(exec_buf, index + 2);
				result[i] = new byte[len];			
  				System.arraycopy(exec_buf, index + 4, result[i], 0, len);

				index += (4 + len);
	    	}
	    	else 
	    	{ /* Tiny arg */
  				len = exec_buf[index + 1] & 0xff;
				result[i] = new byte[len];			
  				System.arraycopy(exec_buf, index + 2, result[i], 0, len);

				index += (2 + len);
	    	}
		}

    	return result;
	}


	/**
	* Error checking. Throws a DlpException when the length of the array is shorter than the
	* specified minimum length.
	*/
	private static void _expect(byte[] b, int minlen)
		throws DlpException
	{
		if (b == null)
		{
			if (minlen == 0)
				return;
			else
				throw new DlpException(1);
			
		}
		if (b.length < minlen)
			throw new DlpException(1);
	}


	/**
	* All DLP functions that return records from databases use an identical
	* response structure. This structure is unpacked by this method.
	*/
	private static Record _unpackRecord(byte[] result, DatabaseImpl dbimpl)
		throws DlpException
	{
		_expect(result, 10);

    	RecordID rsp_id = new RecordID(PackUtils.getLong(result, 0)); 
    	int index = PackUtils.getShort(result, 4);
    	int size = PackUtils.getShort(result, 6);
    	int attr = result[8] & 0xff;
    	int category = result[9] & 0xff;

		byte[] content = new byte[result.length - 10];
    	System.arraycopy(result, 10, content, 0, content.length);

		return dbimpl.createRecord(content, rsp_id, index, attr, category);
	}


	/**
	* All DLP functions that return resources from databases use an identical
	* response structure. This structure is unpacked by this method.
	*/
	private static Resource _unpackResource(byte[] result, DatabaseImpl dbimpl)
		throws DlpException
	{
		_expect(result, 10);
	
		Char4 type = new Char4(PackUtils.getLong(result, 0));
		int id = PackUtils.getShort(result, 4);
		int index = PackUtils.getShort(result, 6);
		int size = PackUtils.getShort(result, 8);
		
		byte[] content = new byte[size];
		System.arraycopy(result, 10, content, 0, size);
		
		return dbimpl.createResource(content, type, id, index);
	}


	/**
	* Extract a DBInfo structure at a given offset within the array.
	*/
	public DBInfo _extractDBInfo(byte[] result, int offset, int cardno)
	{
		int miscflags = 0;

		if (_version >= com.syncbuilder.sync.Dlp.DL_1_1) // PalmOS 2.0+ has additional flags 
		{
		    miscflags = result[1 + offset] & 0xff;
		}

    	int flags = PackUtils.getShort(result, 2 + offset);

	    Char4 type = new Char4((int)(PackUtils.getLong(result, 4 + offset)));
    	Char4 creator = new Char4((int)(PackUtils.getLong(result, 8 + offset)));

	    int version = PackUtils.getShort(result, 12 + offset);
	    int modnum = PackUtils.getLong(result, 14 + offset);

	    java.util.Date create_date = _getDate(result, 18 + offset);
	    java.util.Date modify_date = _getDate(result, 26 + offset);
	    java.util.Date backup_date = _getDate(result, 34 + offset);
	    int index = PackUtils.getShort(result, 42 + offset);

		String name = PackUtils.getZeroTermString(result, 44 + offset, null);

		return new DBInfo(name, type, creator, flags, miscflags, version, modnum,
		index, cardno, create_date, modify_date, backup_date);
	}


	/**
	* This conversion function is strictly for use within the DLP layer.
	* This particular date/time format does not occur anywhere else within the
	* Palm OS or its communications. 
	*/
	private static java.util.Date _getDate(final byte[] data, int offset)
		throws ArrayIndexOutOfBoundsException
	{
    	int sec   = data[offset + 6] & 0xff;
    	int min   = data[offset + 5] & 0xff;
    	int hour  = data[offset + 4] & 0xff;
    	int mday  = data[offset + 3] & 0xff;
    	int mon   = (data[offset + 2] & 0xff) - 1;
    	int year  = (((data[offset + 0] & 0xff)<<8) | (data[offset + 1] & 0xff)) - 1900;

		if (year == -1900)
			return null;
		else
	    	return new java.util.Date(year, mon, mday, hour, min, sec);
	}

	
	/**
	* This conversion function is strictly for use within the DLP layer.
	* This particular date/time format does not occur anywhere else within the
	* Palm OS or its communications. 
	*/
	private static void _setDate(byte[] data, int offset, java.util.Date date)
		throws ArrayIndexOutOfBoundsException
	{
		if (date == null)
		{
			for(int i = 0; i < 8; i++)
			{
				data[offset + i] = (byte)0;
			}
		
			return;
		}

        data[offset + 7] = 0; // padding
        data[offset + 6] = (byte)(date.getSeconds());
        data[offset + 5] = (byte)(date.getMinutes());
        data[offset + 4] = (byte)(date.getHours());
        data[offset + 3] = (byte)(date.getDate());
        data[offset + 2] = (byte)(date.getMonth() + 1);

        int y = date.getYear() + 1900;
        data[offset + 0] = (byte)((y >> 8) & 0xff);
        data[offset + 1] = (byte)((y >> 0) & 0xff);
	}


	private final java.io.InputStream  _in;
	private final java.io.OutputStream _out;
	private final int _version;  // Implement which version of the DL protocol

	private final static int DLP_BUF_SIZE = 0xffff; // Is the 64KB limit still there in the PalmIII ???


	private final static int FIRSTARGID = 0x20;     // First possible argument ID
}
