/* com/syncbuilder/internal/sync/NativeSocketImpl.java:
*
* Copyright (C) 1998, 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: NativeSocketImpl.java,v 1.5 1999/03/18 22:36:24 christ Exp $ */


package com.syncbuilder.internal.sync;

import java.io.*;

/**
* The implementation details of the Sockets which use the pilot-link native sockets.<br>
* This is to be considered an intermediary solution until the Pure Java implementations
* are ready for prime-time.
*
* @author Tilo Christ
*/
public class NativeSocketImpl
	extends com.syncbuilder.sync.SocketImpl
{
	public boolean isOpen()
	{
		return (_sdesc != 0);
	}


	synchronized public int getVersion() 
		throws IOException
    { 
		if (_version_determined)
			return _version;
	
		int comm_version = _pi_version(_sdesc);
		
		if (comm_version < _COMM_VERSION_2_0)
		{
			// commVersion is 1.0. DLP version will be 1.0.
			_version = com.syncbuilder.sync.Dlp.DL_1_0;
		}
		else
		{
			// commVersion is 2.0. We have to probe for correct DLP-version (1.1 or 1.2)

			try
			{
				com.syncbuilder.internal.sync.DlpImpl dlp_impl = new com.syncbuilder.internal.sync.DlpImpl(getInputStream(), getOutputStream());
				com.syncbuilder.device.SysInfo sysinfo = dlp_impl.readSysInfoVer();

				_version = sysinfo.getDLPVersion();

				if (_DEBUG)
					System.err.println("Probed DL-version is: " + _version);
			}
			catch(Exception e)
			{
				if (_DEBUG)
					e.printStackTrace();
				_version = com.syncbuilder.sync.Dlp.DL_1_1;

				if (_DEBUG)
					System.err.println("Probing failed. Using DL 1.1");
			}
			
		}
		
		_version_determined = true;
		return _version;
	}

	
	/**
	* "Tickle" the PalmPilot in order to keep the connection alive. 
	* This can be automated through startWatchdog.
	*
	* @see #startWatchdog
	*/
    synchronized public void tickle() 
		throws IOException
    { 
		_pi_tickle(_sdesc); 
	}

	
	/**
	* Tickle the PalmPilot automatically in specified intervals
	*
	* @see #tickle
	*/
	synchronized public void startWatchdog(int interval) 
		throws IOException
	{
		_pi_watchdog(_sdesc, interval); 
	}


	/**
	* Read a specified number of bytes into the specified array.<br>
	*/
	public synchronized int read(byte[] data, int len) 
		throws IOException
	{ 
		return _pi_read(_sdesc, data, len); 
	}
	
	/**
	* Write a specified number of bytes from the specified array.
	*/
	public synchronized int write(byte[] data, int len) 
		throws IOException
	{ 
		return _pi_write(_sdesc, data, len); 
	}


	/**
	* Fill in the socket-descriptor of the supplied SocketImpl.
	* This implies that the SocketImpl is a NativeSocketImpl.
	*/    
	synchronized public void accept(com.syncbuilder.sync.SocketImpl impl, int timeout) 
		throws IOException
    { 
		if (!(impl instanceof com.syncbuilder.internal.sync.NativeSocketImpl))
			throw new IllegalArgumentException("impl is not a NativeSocketImpl");

		((NativeSocketImpl)impl)._sdesc = _pi_accept(_sdesc, timeout); 
	}

	/**
	* Fill in the socket-descriptor of the supplied SocketImpl.
	* This implies that the SocketImpl is a NativeSocketImpl.
	*/    
	synchronized public void accept(com.syncbuilder.sync.SocketImpl impl) 
		throws IOException
    { 
		accept(impl, 0);
	}

	synchronized public void bind(String device) 
		throws IOException
	{ 
		_pi_bind(_sdesc, device); 
	}

    synchronized public void listen(int backlog) 
		throws IOException
    { 
		_pi_listen(_sdesc, backlog); 
	}


	/** Return the maximum number of simultaneous connections which can be handled by this kind of socket */
	public int getMaxConnections()
	{
		return 1;
	}


	synchronized public void close() 
		throws IOException, com.syncbuilder.sync.DlpException 
	{ 
		/* this method must be idempotent */
		if (this._sdesc != 0)
		{
			_pi_noendsync(_sdesc);
	    	_pi_close(_sdesc); 
		}
			
		_sdesc = 0;
	}

	/** Switch the SocketImpl into ServerSocket-mode */
	public void create()
		throws IOException
	{
		_sdesc = _pi_socket(0x51, 0x10, 0x51);
		_pi_noendsync(_sdesc); // A ServerSocket never needs the end-of-hotsync sequence
	}


	/**
	* Only NativeSocketImplFactory shall instantiate this class.
	*/
	public NativeSocketImpl()
	{
	}
	
	
	// These constants refer to the version of the CMP protocol.
	// There is not neccessarily any relation to the version of the DLP protocol!!!
	private final static int _COMM_VERSION_1_0 = 0x0100;
	private final static int _COMM_VERSION_2_0 = 0x0101;
	
	
	// This descriptor is only meaningful to the native pilot-link code.
	private int _sdesc = 0;

	private static  native void _pi_close(int socket) throws IOException;
	private static  native int _pi_socket(int domain, int type, int protocol) throws IOException;
	private static  native void _pi_bind(int socket, String device) throws IOException;
	private static  native void _pi_listen(int socket, int backlog) throws IOException;
	private static  native int _pi_accept(int socket, int timeout) throws IOException;
	private static  native int _pi_version(int socket) throws IOException;
	private static  native void _pi_tickle(int socket) throws IOException;
	private static  native void _pi_watchdog(int socket, int interval) throws IOException;
	private static  native int _pi_read(int socket, byte[] data, int len) throws IOException;
	private static  native int _pi_write(int socket, byte[] data, int len) throws IOException;
	private static  native void _pi_noendsync(int socket);

	private int _version;
	private boolean _version_determined = false;

	private final static boolean _DEBUG = false;

	static 
	{
		System.loadLibrary("JavaPisock");
	}
}
