/* com/syncbuilder/storage/DBInfo.java:
*
* Copyright (C) 1998, 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
/* $Id: DBInfo.java,v 1.7 1999/02/13 20:33:39 christ Exp $ */

package com.syncbuilder.storage;

import java.util.Date;
import com.syncbuilder.internal.core.constants;

/**
* Meta-information of a database.
*
* @author Tilo Christ
*/
public class DBInfo 
{
	/**
	* Construct a new DBInfo. 
	*/
	public DBInfo(String name, Char4 type, Char4 creator, int flags, int miscflags, int version, int modnum,
		int index, int card, Date create_date, Date modify_date, Date backup_date)
	{
		this._name = name;
		this._type = type;
		this._creator = creator;
		this._flags = flags;
		this._miscflags = miscflags;
		this._version = version;
		this._modnum = modnum;
		this._index = index;
		this._card = card;
		this._create_date = create_date;
		this._modify_date = modify_date;
		this._backup_date = backup_date;
	}


	/**
	* Return the name of the database.
	*/
	public String getName()
	{
		return _name;
	}
	
	
	/**
	* Return the type of the database.
	*/
	public Char4 getType()
	{
		return _type;
	}
	
	
	/**
	* Return the creator ID of the database.
	*/
	public Char4 getCreator()
	{
		return _creator;
	}
	

	/**
	* Return the version of the database. This is typically also the version number
	* of the application which is contained within that database.
	*/	
	public int getVersion()
	{
		return _version;
	}
	
		
	/**
	* Return the modification number of the database.
	*/	
	public int getModificationNumber()
	{
		return _modnum;
	}
	
		
	/**
	* Return the index of the database.
	*
	* @return the index or <code>UNKNOWN_DB_INDEX</code>.
	*/	
	public int getIndex()
	{
		return _index;
	}
	
		
	/**
	* Return the number of the card, on which the database is located.
	*/	
	public int getCard()
	{
		return _card;
	}
	
		
	public int getFlags()
	{
		return _flags;
	}


	public int getMiscFlags()
	{
		return _miscflags;
	}


	public Date getCreationDate()
	{
		return _create_date;
	}
	

	public Date getModificationDate()
	{
		return _modify_date;
	}
	

	public Date getBackupDate()
	{
		return _backup_date;
	}
	

	/**
	* Is this database read-only?
	*/
	public boolean isReadOnly()
	{
		return ((_flags & FLAG_READONLY) != 0);
	}


	/**
	* Is this a resource database?
	* @return true = resources, false = records
	*/
	public boolean isResource()
	{
		return ((_flags & FLAG_RESOURCE) != 0);
	}


	/**
	* Is this database open?
	*/
	public boolean isOpen()
	{
		return ((_flags & FLAG_OPEN) != 0);
	}


	/**
	* Is this a database scheduled for backup?
	*/
	public boolean isBackup()
	{
		return ((_flags & FLAG_BACKUP) != 0);
	}


	/**
	* Has the AppInfo of this database been changed?
	*/
	public boolean isAppInfoDirty()
	{
		return ((_flags & FLAG_APPINFO_DIRTY) != 0);
	}


	/**
	* May this database be overwritten by newer databases, even though it is currently opened?
	*/
	public boolean isNewer()
	{
		return ((_flags & FLAG_NEWER) != 0);
	}


	/**
	* Shall the device be reset after this database has been installed?
	*/
	public boolean isReset()
	{
		return ((_flags & FLAG_RESET) != 0);
	}


	/**
	* Shall the copying (IR beaming) of this database be prevented?
	*/
	public boolean isCopyPrevention()
	{
		return ((_flags & FLAG_COPY_PREVENTION) != 0);
	}


	/**
	* Shall this database be excluded from all synchronization?
	*/
	public boolean isExcludeFromSync()
	{
		return ((_miscflags & MISCFLAG_EXCLUDE_FROM_SYNC) != 0);
	}


	/**
	* Is this database located in RAM?
	*/
	public boolean isRamBased()
	{
		return ((_miscflags & MISCFLAG_RAM_BASED) != 0);
	}


	public String toString() 
	{
		return "<com.syncbuilder.storage.DBInfo Name '" + _name + "', Version " + _version +
			", Read Only=" + isReadOnly() + ", Resource=" + isResource() +
			", Open=" + isOpen() + ", Backup=" + isBackup() +
			", AppInfo Dirty=" + isAppInfoDirty() + ", Newer=" + isNewer() +
			", Reset=" + isReset() + ", Exclude from Sync=" + isExcludeFromSync() +
			", RAM Based=" + isRamBased() +
			", Creator " + _creator.toString() +
			", Type " + ((_type==null) ? "unknown" : _type.toString()) + ", Modification# " + _modnum +
			", Create Date " + ((_create_date==null) ? "unknown" : _create_date.toString()) +
			", Modify Date " + ((_modify_date==null) ? "unknown" : _modify_date.toString()) +
			", Backup Date " + ((_backup_date==null) ? "unknown" : _backup_date.toString()) +
			", Index " + ((_index == UNKNOWN_DB_INDEX) ? "unknown" : Integer.toString(_index)) +
			", Card " + _card + ">";
	}


	public final static int FLAG_READONLY = constants.dlpDBFlagReadOnly;
	public final static int FLAG_RESOURCE = constants.dlpDBFlagResource;
	public final static int FLAG_OPEN = constants.dlpDBFlagOpen;
	public final static int FLAG_BACKUP = constants.dlpDBFlagBackup;
	public final static int FLAG_APPINFO_DIRTY = constants.dlpDBFlagAppInfoDirty;
	public final static int FLAG_NEWER = constants.dlpDBFlagNewer;
	public final static int FLAG_RESET = constants.dlpDBFlagReset;
	public final static int FLAG_COPY_PREVENTION = 0x0040; // dlpDBFlagCopyPrevention

	public final static int MISCFLAG_EXCLUDE_FROM_SYNC = constants.dlpDBMiscFlagExcludeFromSync;
	public final static int MISCFLAG_RAM_BASED = 0x40;  // dlpDbInfoMiscFlagRamBased

	/** The real value of the index is unknown */
	public final static int UNKNOWN_DB_INDEX = 0xFFFF;  // dlpDbInfoUnknownDbIndex


	private String _name;
	private Char4 _creator;
	private Char4 _type;

	private Date _create_date;
	private Date _modify_date; 
	private Date _backup_date;

	private int _version;
	private int _modnum;
	private int _index;
	private int _card;
	
	private int _flags;
	private int _miscflags;
}

