/* com/syncbuilder/storage/Database.java:
*
* Copyright (C) 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Database.java,v 1.2 1999/02/18 09:45:36 christ Exp $ */

package com.syncbuilder.storage;

import com.syncbuilder.internal.core.constants;

/** 
 * This interface is implemented by all classes which represent a database.
 * This includes databases on the device, as well as those in any other storage.
 *
 * @author Tilo Christ
 */
public interface Database
{
//=============================== AppBlock ====================================

	/**
	* Get a new instance of an AppBlock that matches the Database.
	*/
	public AppBlock getAppBlock() 
		throws DatabaseException;

	/**
	* Associate the AppBlock with the Database.
	*/
	public void setAppBlock(AppBlock appblock) 
		throws DatabaseException;

	/**
	* Create a new instance of an AppBlock that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create an AppBlock.
	*/
	public AppBlock createAppBlock();

//=============================== SortBlock ====================================

	/**
	* Get a new instance of a SortBlock that matches the Database.
	*/
	public SortBlock getSortBlock() 
		throws DatabaseException;

	/**
	* Associate the SortBlock with the Database.
	*/
	public void setSortBlock(SortBlock sortblock) 
		throws DatabaseException;

	/**
	* Create a new instance of a SortBlock that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create a SortBlock.
	*/
	public SortBlock createSortBlock();

//=============================== Record ====================================

	/**
	* Retrieve the Record with the specified index from the Database.
	* 
	* @param index the index of the Record
	* @return a specialization of the Record class which matches the creator of the Database, or <code>null</code>, if the specified Record could not be found.
	*/
	public Record getRecord(int index) 
		throws DatabaseException;

	/**
	* Retrieve the Record with the specified UID from the Database.
	* 
	* @param uid the UID of the Record
	* @return a specialization of the Record class which matches the creator of the Database, or <code>null</code>, if the specified Record could not be found.
	*/
	public Record getRecord(RecordID uid) 
		throws DatabaseException;

	/**
	* Retrieve the next Record within the specified category from the Database.
	* The database keeps an internal 'next record' index pointer, which it updates automatically.
	* 
	* @param category the category from which the Record shall be taken
	* @return a specialization of the Record class which matches the creator of the Database, or <code>null</code>, if the specified Record could not be found.
	* @see #resetNext
	*/
	public Record getNextRecord(int category) 
		throws DatabaseException;

	/**
	* Retrieve the next modified Record from the Database.
	* The database keeps an internal 'next record' index pointer, which it updates automatically.
	* 
	* @return a specialization of the Record class which matches the creator of the Database, or <code>null</code>, if the specified Record could not be found.
	* @see #resetNext
	*/
	public Record getNextModifiedRecord() 
		throws DatabaseException;

	/**
	* Retrieve the next modified Record within the specified category from the Database.
	* The database keeps an internal 'next record' index pointer, which it updates automatically.
	* 
	* @param category the category from which the Record shall be taken
	* @return a specialization of the Record class which matches the creator of the Database, or <code>null</code>, if the specified Record could not be found.
	* @see #resetNext
	*/
	public Record getNextModifiedRecord(int category) 
		throws DatabaseException;

	/**
	* Resets the 'next record' index to the beginning
	*/
	public void resetNext() 
		throws DatabaseException;

	/**
	* Stores a record in the database.
	* If the Record had an unspecified RecordID before,
	* then a new valid RecordID will be assigned to it.
	*/
	public void putRecord(Record record) 
		throws DatabaseException;

	/**
	* Create a new Record with an unspecified RecordID (0).
	* <p>
	* Note: Avoid using the new-operator in order to create a Record.
	*/
	public Record createRecord();

	/**
	* Create a new Record with the specified RecordID.
	* <p>
	* Note: Avoid using the new-operator in order to create a Record.
	*
	* @param id the unique ID of the record
	*/
	public Record createRecord(RecordID id);

	/**
	* Create a new Record with the specified RecordID.
	* <p>
	* Note: Avoid using the new-operator in order to create a Record.
	*
	* @param contents the raw contents of the Record
	* @param id the unique ID of the record
	* @param index the index of the record within the database
	* @param attr the attributes of the record
	* @param cat the category of the record
	*/
	public Record createRecord(byte[] contents, RecordID id, int index, int attr, int cat);

	/**
	* Delete the record with the specified UID from the database.
	* 
	* @param uid the UID (not the index!) of the record
	*/
	public void deleteRecord(RecordID uid) 
		throws DatabaseException;

	/**
	* Delete ALL records in the database.
	* <br><b>Use with caution!</b>
	*/
	public void deleteAllRecords() 
		throws DatabaseException;

	/**
	* Delete all records from a specified category
	*
	* @param category the category
	*/
	public void deleteCategory(int category)
		throws DatabaseException;

	public void moveCategory(int from, int to) 
		throws DatabaseException;

	/**
	* Return the UIDs of all records in the DB, in an unsorted fashion.
	*/
	public RecordID[] getRecordIDs()	
		throws DatabaseException;

	/**
	* Return the UIDs of all records in the DB. If you ask for a sorted list
	* of UIDs, then the application on the Palm device will be asked to sort the records
	* before their UIDs are returned.
	*/
	public RecordID[] getRecordIDs(boolean sort) 	
		throws DatabaseException;

	/**
	* Return the UIDs of selected records in the DB. If you ask for a sorted list
	* of UIDs, then the application on the Palm device will be asked to sort the records
	* before their UIDs are returned.
	*/
	public RecordID[] getRecordIDs(boolean sort, int start)		
		throws DatabaseException;

	/**
	* Return the UIDs of selected records in the DB. If you ask for a sorted list
	* of UIDs, then the application on the Palm device will be asked to sort the records
	* before their UIDs are returned.
	*/
	public RecordID[] getRecordIDs(boolean sort, int start, int max)
		throws DatabaseException;

	/**
	* Deletes all records which are marked as archived or deleted
	*/
	public void purgeRecords() 
		throws DatabaseException;



//=============================== Resource ====================================

	/**
	* Get a resource through the specified index.
	*/
	public Resource getResource(int index) 
		throws DatabaseException;

	/**
	* Get a resource through the type and id.
	*/
	public Resource getResource(Char4 type, int id) 
		throws DatabaseException;

	/**
	* Store a resource in the database.
	*/
	public void putResource(Resource resource) 
		throws DatabaseException;

	/**
	* Create a new instance of a Resource that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create an Resource.
	*/
	public Resource createResource();

	/**
	* Create a new instance of a Resource that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create an Resource.
	*
	* @param type the type of the Resource
	* @param id the id of the Resource
	*/
	public Resource createResource(Char4 type, int id);

	/**
	* Create a new instance of a Resource that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create an Resource.
	*
	* @param contents the raw contents of the Resource
	* @param type the type of the Resource
	* @param id the id of the Resource
	* @param index the index of the Resource with in the database
	*/
	public Resource createResource(byte[] contents, Char4 type, int id, int index);

	/**
	* Delete a single resource from the database.
	*
	* @param type the type of the Resource
	* @param id the id of the Resource
	*/
	public void deleteResource(Char4 type, int id) 
		throws DatabaseException;

	/**
	* Delete ALL resources.
	* <br><b>Use with caution!</b>
	*/
	public void deleteAllResources() 
		throws DatabaseException;


//=============================== Preference ====================================

	
	/**
	* Get one of the preferences which are associated with the same app as the database.
	* The Pref is taken from the saved preferences.
	*
	* @param id the id of the preference
	*/
	public Pref getPref(int id)
	  throws DatabaseException, NoCreatorException;

	/**
	* Get one of the preferences which are associated with the same app as the database.
    *
	* @param id the id of the preference
	* @param backed_up Shall the Pref be taken from the saved (true) or from the unsaved (false) preferences?
	*/
	public Pref getPref(int id, boolean backed_up)
		throws DatabaseException, NoCreatorException;

	/**
	* Store a Pref which is associated with the same app as the database to the database.
	* Since the Creator ID is taken from the Pref object and not from the database object,
	* this method will also happily write preferences which belong to a different application.
	*/
	public void setPref(Pref pref)
		throws DatabaseException;

	/**
	* Create a new instance of a Pref that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create a Pref.
    *
	* @param id the id of the preference
	*/
	public Pref createPref(int id);

	/**
	* Create a new instance of a Pref that matches the Database.
	* <p>
	* Note: Avoid using the new-operator in order to create a Pref.
    *
	* @param id the id of the preference
	* @param version the version number of the application
	* @param backed_up Shall the Pref be taken from the saved (true) or from the unsaved (false) preferences?
	*/
	public Pref createPref(int id, int version, boolean backed_up);


//=============================== Miscellaneous ====================================

	/**
	* Closes the database. 
	* The Database-object shall no longer be used after the invocation of this method!
	*/
	public void close() 
		throws DatabaseException;

	/**
	* For record databases, reset all dirty flags. For both record and
	* resource databases, set the last sync time to now. 
	*/
	public void resetFlags() 
		throws DatabaseException;

	/**
	* Return the number of records in the Database.
	*/
	public int getRecordCount() 
		throws DatabaseException;

	/**
	* Return meta-information about the Database.
	*/
	public DBInfo getDBInfo()
		throws DatabaseException;


	/**
	* Associate a complete set of meta-information with the Database.
	*/
	public void setDBInfo(DBInfo dbinfo)
		throws DatabaseException;


	/* 
	  These constants are used to describe the mode in which
	  the database is to be opened. Not all implementations will
	  honour these flags.
	*/
	/** Opened for read access. Not all implementations will honour this flag. */
	public final static int OPEN_READ       = constants.dlpOpenRead;
	/** Opened for write access. Not all implementations will honour this flag. */
	public final static int OPEN_WRITE      = constants.dlpOpenWrite;
	/** Opened for exclusive access. Not all implementations will honour this flag. */
	public final static int OPEN_EXCLUSIVE  = constants.dlpOpenExclusive;
	
	/** Opened with default settings (read/write, exclusive). Not all implementations will honour this flag. */
	public final static int OPEN_DEFAULT = OPEN_READ|OPEN_WRITE|OPEN_EXCLUSIVE;
}

